/**
 * Copyright (C) cedarsoft GmbH.
 *
 * Licensed under the GNU General Public License version 3 (the "License")
 * with Classpath Exception; you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *         http://www.cedarsoft.org/gpl3ce
 *         (GPL 3 with Classpath Exception)
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3 only, as
 * published by the Free Software Foundation. cedarsoft GmbH designates this
 * particular file as subject to the "Classpath" exception as provided
 * by cedarsoft GmbH in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 3 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 3 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact cedarsoft GmbH, 72810 Gomaringen, Germany,
 * or visit www.cedarsoft.com if you need additional information or
 * have any questions.
 */

package com.cedarsoft.gdao;

import com.cedarsoft.CommitableObjectAccess;
import org.jetbrains.annotations.NotNull;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.EmptyResultDataAccessException;

import java.util.List;

/**
 * A generic dao. This implementation expects a Long as PK
 *
 * @param <T> the type this dao can manage
 */
public interface GenericDao<T> extends CommitableObjectAccess<T> {
  /**
   * Saves the newInstance object into database
   */
  @NotNull
          <LT extends T> Long save( @NotNull LT newInstance ) throws DataAccessException;

  /**
   * Save changes made to a persistent object.
   */
  <LT extends T> void update( @NotNull LT transientObject ) throws DataAccessException;

  /**
   * Stores the changes to a persistent object or saves it
   *
   * @param object the object
   */
  <LT extends T> void saveOrUpdate( @NotNull LT object ) throws DataAccessException;

  /**
   * Remove an object from persistent storage in the database
   */
  <LT extends T> void delete( @NotNull LT persistentObject ) throws DataAccessException;

  /**
   * Retrieve an object that was previously persisted to the database using
   * the indicated id as primary key
   *
   * @param id the id of the bean that is loaded
   */
  @NotNull
  T findById( @NotNull Long id ) throws EmptyResultDataAccessException;

  /**
   * Finds an object.
   *
   * @param finder the selector
   * @return the selected objects
   *
   * @throws EmptyResultDataAccessException if nothing has been found
   */
  @NotNull
          <R> R find( @NotNull Finder<R> finder ) throws EmptyResultDataAccessException;

  /**
   * Selects all objects
   *
   * @return all objects
   */
  @NotNull
  List<? extends T> findAll() throws DataAccessException;

  /**
   * Returns the count
   *
   * @return the count
   */
  int getCount() throws DataAccessException;

  /**
   * Shuts the dao down
   */
  void shutdown();
}
