/**
 * Copyright (C) cedarsoft GmbH.
 *
 * Licensed under the GNU General Public License version 3 (the "License")
 * with Classpath Exception; you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *         http://www.cedarsoft.org/gpl3ce
 *         (GPL 3 with Classpath Exception)
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3 only, as
 * published by the Free Software Foundation. cedarsoft GmbH designates this
 * particular file as subject to the "Classpath" exception as provided
 * by cedarsoft GmbH in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 3 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 3 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact cedarsoft GmbH, 72810 Gomaringen, Germany,
 * or visit www.cedarsoft.com if you need additional information or
 * have any questions.
 */

package com.cedarsoft.gdao;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.Override;

/**
 * A generic service that extends the behaviour of simple DAOs with transaction support.
 */
public interface GenericService<T> extends GenericDao<T> {
  /**
   * Performs an action within an transaction.
   *
   * @param callback the callback that is called within a transaction
   * @return the return value (optional)
   */
  @Nullable
      <R> R perform( @NotNull ServiceCallback<T, R> callback );


  /**
   * A callback that can be used to be invoked within a service (in a transaction).
   */
  interface ServiceCallback<T, R> {
    /**
     * Perform the action
     *
     * @param service the service
     * @return the return value (may be null)
     */
    @Nullable
    R perform( @NotNull GenericService<T> service );
  }

  /**
   * Abstract base class for service callbacks without return value.
   */
  abstract class ServiceCallbackWithoutReturnValue<T> implements ServiceCallback<T, Object> {
    @Override
    @Nullable
    public Object perform( @NotNull GenericService<T> service ) {
      performVoid( service );
      return null;
    }

    /**
     * Perform the action without any return value
     *
     * @param service the service
     */
    protected abstract void performVoid( @NotNull GenericService<T> service );
  }
}

