/**
 * Copyright (C) cedarsoft GmbH.
 *
 * Licensed under the GNU General Public License version 3 (the "License")
 * with Classpath Exception; you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *         http://www.cedarsoft.org/gpl3ce
 *         (GPL 3 with Classpath Exception)
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3 only, as
 * published by the Free Software Foundation. cedarsoft GmbH designates this
 * particular file as subject to the "Classpath" exception as provided
 * by cedarsoft GmbH in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 3 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 3 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact cedarsoft GmbH, 72810 Gomaringen, Germany,
 * or visit www.cedarsoft.com if you need additional information or
 * have any questions.
 */

package com.cedarsoft.gdao.async;

import com.cedarsoft.async.AsyncCallSupport;
import com.cedarsoft.async.CallbackCaller;
import com.cedarsoft.cache.Cache;
import com.cedarsoft.cache.HashedCache;
import com.cedarsoft.gdao.GenericService;
import com.cedarsoft.gdao.GenericServiceManager;
import org.jetbrains.annotations.NotNull;

import java.lang.Override;
import java.util.Iterator;

/**
 *
 */
public class AsynchronousServiceManager implements GenericServiceManager {
  @NotNull
  private final AsyncCallSupport<AsyncServiceCreator<?>> asyncCallSupport = new AsyncCallSupport<AsyncServiceCreator<?>>();

  @SuppressWarnings( {"MismatchedQueryAndUpdateOfCollection"} )
  @NotNull
  private final Cache<Class<Object>, AsynchronousService<Object>> serviceCache = new HashedCache<Class<Object>, AsynchronousService<Object>>(
          new Cache.Factory<Class<Object>, AsynchronousService<Object>>() {
            @Override
            @NotNull
            public AsynchronousService<Object> create( @NotNull Class<Object> key ) {
              return asyncCallSupport.invoke( new AsyncServiceCreator<Object>( key ) );
            }
          }
  );

  public AsynchronousServiceManager( @NotNull final String description, @NotNull final GenericServiceManager delegatingManager ) {
    asyncCallSupport.initializeWorker( new CallbackCaller<AsyncServiceCreator<?>>() {
      @Override
      @NotNull
      public String getDescription() {
        return description;
      }

      @Override
      public Object call( @NotNull AsyncServiceCreator<?> callback ) throws Exception {
        return callback.execute( delegatingManager );
      }
    } );
  }

  @Override
  @NotNull
  public <T> GenericService<T> getService( @NotNull Class<T> type ) {
    return ( GenericService<T> ) serviceCache.get( type );
  }

  public void shutdown() {
    for ( Iterator<AsynchronousService<Object>> it = serviceCache.values().iterator(); it.hasNext(); ) {
      AsynchronousService<Object> service = it.next();
      service.shutdown();
      it.remove();
    }
    asyncCallSupport.shutdown();
  }

  private static class AsyncServiceCreator<T> {
    @NotNull
    private final Class<T> key;

    private AsyncServiceCreator( @NotNull Class<T> key ) {
      this.key = key;
    }

    @NotNull
    public AsynchronousService<T> execute( @NotNull GenericServiceManager delegatingManager ) {
      GenericService<T> service = delegatingManager.getService( key );
      AsynchronousService<T> asynchronousService = new AsynchronousService<T>();
      asynchronousService.initializeDelegatingService( service );
      return asynchronousService;
    }
  }
}
